// AISearchAgentAlphabeta.cpp: Implementierung der Klasse AISearchAgentAlphabeta.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AISearchAgentAlphabeta.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

const int AISearchAgentAlphabeta::SearchDepth = 5;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/

AISearchAgentAlphabeta::AISearchAgentAlphabeta() : AISearchAgent()
{

}

AISearchAgentAlphabeta::~AISearchAgentAlphabeta()
{

}


/***************************************************************************
 * Operations
 **************************************************************************/

void AISearchAgentAlphabeta::PickBestMove( Board* theBoard, Move* bestMove )
{
	// Store the identity of the moving side, so that we can tell Evaluator
	// from whose perspective we need to evaluate positions
	FromWhosePerspective = theBoard->GetCurrentPlayer();
	MoveCounter++;

	// Should we erase the history table?
	if ( ( rand() % 4 ) == 2 )
		CutoffHist.Forget();

	NumRegularNodes = 0; NumQuiescenceNodes = 0;
	NumRegularTTHits = 0; NumQuiescenceTTHits = 0;

	// Find the moves
	Move theMove;
	MoveListGenerator movegen;
	movegen.ComputeLegalMoves( theBoard );
	CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	// The following code blocks look a lot like the MAX node case from
	// AISearchAgent.Alphabeta, with an added twist: we need to store the
	// actual best move, and not only pass around its minimax value
	int bestSoFar = ALPHABETA_MINVAL;
	Board newBoard;
	Move *mov;
	int currentAlpha = ALPHABETA_MINVAL;

	// Loop on all pseudo-legal moves
	while( ( mov = movegen.Next() ) != NULL ) {
		newBoard.Clone( theBoard );
		newBoard.ApplyMove( mov );
		int movScore = AlphaBeta( MINNODE, &newBoard, SearchDepth, currentAlpha, ALPHABETA_MAXVAL );
		if ( movScore == ALPHABETA_ILLEGAL )
			continue;

		currentAlpha = max( currentAlpha, movScore );

		if ( movScore > bestSoFar ) {
			theMove.Set( mov );
			bestSoFar = movScore;
			theMove.MoveEvaluation = movScore;
		}
    }

	// And now, if the best we can do is ALPHABETA_GIVEUP or worse, then it is
	// time to resign...  Unless the opponent was kind wnough to put us in
	// stalemate!
	if ( bestSoFar <= ALPHABETA_GIVEUP ) {
		// Check for a stalemate
		// Stalemate occurs if the player's king is NOT in check, but all of his
		// moves are illegal.
		// First, verify whether we are in check
		newBoard.Clone( theBoard );
		MoveListGenerator secondary;
		newBoard.SwitchSides();
		if ( secondary.ComputeLegalMoves( &newBoard ) ) {
			// Then, we are not in check and may continue our efforts.
			// We must now examine all possible moves; if at least one resuls in
			// a legal position, there is no stalemate and we must assume that
			// we are doomed
			CutoffHist.SortMoveList( &movegen, newBoard.GetCurrentPlayer() );
			movegen.ResetIterator();
			// If we can scan all moves without finding one which results
			// in a legal position, we have a stalemate
			theMove.MoveType = Move::MOVE_STALEMATE;
			theMove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			// Look at the moves
			while( ( mov = movegen.Next() ) != NULL ) {
				newBoard.Clone( theBoard );
				newBoard.ApplyMove( mov );
				if ( secondary.ComputeLegalMoves( &newBoard ) ) {
					theMove.MoveType = Move::MOVE_RESIGN;
				}
			}
		} else {
			// We're in check and our best hope is GIVEUP or worse, so either we are
			// already checkmated or will be soon, without hope of escape
			theMove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			theMove.MoveType = Move::MOVE_RESIGN;
		}
    }

    *Output::cout <<  "MoveScore: " << bestSoFar << endl ;
    *Output::cout <<  "  --> Transposition Table hits for regular nodes: " ;
    *Output::cout << NumRegularTTHits << " of " << NumRegularNodes << endl;
    *Output::cout <<"  --> Transposition Table hits for quiescence nodes: " ;
    *Output::cout << NumQuiescenceTTHits << " of " << NumQuiescenceNodes << endl;
	
	bestMove->Set( &theMove );

	return;
}
