// AISearchAgentMTDF.cpp: Implementierung der Klasse AISearchAgentMTDF.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AISearchAgentMTDF.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

const int AISearchAgentMTDF::MaxSearchDepth		=      15;
const int AISearchAgentMTDF::MaxSearchSize		=  999999;
const int AISearchAgentMTDF::MaxSearchTime		=   10000;
const int AISearchAgentMTDF::MaxSearchIterTime	=    2500;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/

AISearchAgentMTDF::AISearchAgentMTDF( OpeningBook* ref ) : AISearchAgent()
{
    Openings = ref;
}

AISearchAgentMTDF::~AISearchAgentMTDF()
{

}

/***************************************************************************
 * Operations
 **************************************************************************/

void AISearchAgentMTDF::PickBestMove( Board* theBoard, Move* bestMove )
{

	// First things first: look in the Opening Book, and if it contains a
	// move for this position, don't search anything
	MoveCounter++;
	Move* Mov;
	Timer theTimer;
	Timer iterTimer;


	Mov = Openings->Query( theBoard );
	if ( Mov != NULL ) {
		bestMove->Set( Mov );
		return;
	}

	// Store the identity of the moving side, so that we can tell Evaluator
	// from whose perspective we need to evaluate positions
	FromWhosePerspective = theBoard->GetCurrentPlayer();

	// Should we erase the history table?
	if ( theBoard->MoveNo % 6 == 2 ) 
		CutoffHist.Forget();

	// Begin search.  The search's maximum depth is determined on the fly,
	// according to how much effort has been spent; if it's possible to search
	// to depth 8 in 5 seconds, then by all means, do it!
	int bestGuess = 0;
	int iterdepth = 1;
	int NumAllNodes = 0;

	theTimer.Start();

	while( true ) {
		// Searching to depth 1 is not very effective, so we begin at 2
		iterdepth++;

		// Compute efficiency statistics
		
		NumRegularNodes   = 0; NumQuiescenceNodes   = 0;
		NumRegularTTHits  = 0; NumQuiescenceTTHits  = 0;
		NumRegularCutoffs = 0; NumQuiescenceCutoffs = 0;

		// Look for a move at the current depth

		iterTimer.Start();
		MTDF( theBoard, bestGuess, iterdepth, bestMove );
		iterTimer.Stop();

		NumAllNodes += NumRegularNodes + NumQuiescenceNodes;
   
		bestGuess = bestMove->MoveEvaluation;

		// Feedback!
		*Output::cout <<  "Iteration of depth " << iterdepth << "; best move = ";
		bestMove->Print();
		*Output::cout <<  " ^- AI-movesequence: " << BestMoves::Get(iterdepth, theBoard, &TransTable, &TransTableBlack) << endl;
		
		if ( WinBoardConnection::isActive ) {
			stringstream iostream;
			string smove;
			string sconn;
			string source;
			string dest;
			string sstream;
			string out;

			iostream.clear();
			iostream << iterdepth;
			iostream >> sstream;
			out = sstream + "\t";

			iostream.clear();
			iostream << bestGuess;
			iostream >> sstream;
			out = out + sstream + "\t0\t";

			iostream.clear();
			iostream << ( NumRegularNodes + NumQuiescenceNodes );
			iostream >> sstream;
			out = out + sstream + "\t";

			out = out + BestMoves::Get( iterdepth, theBoard, &TransTable, &TransTableBlack );
					
			WinBoardConnection::getInstance()->send( WinBoardConnection::CMD_THINKING, out );
		
		}

		// Get out if we have searched deep enough
		if ( ( NumRegularNodes + NumQuiescenceNodes ) > MaxSearchSize )
			break;
		if ( iterdepth >= MaxSearchDepth )
			break;
		if ( iterTimer.GetDuration() > MaxSearchIterTime )
			break;
		if ( theTimer.GetDuration() > MaxSearchTime )
			break;
	}

	theTimer.Stop();

	*Output::cout << "AI thought " << theTimer.GetDuration() / 1000 << " seconds! ( " << (NumAllNodes / theTimer.GetDuration()) << " KNodes / sec )" << endl;

	return;
}




int AISearchAgentMTDF::MTDF( Board *theBoard, int target, int depth, Move* bestMove )
{
	int beta;
	Move Mov;
	int currentEstimate = target;
	int upperbound = ALPHABETA_MAXVAL;
	int lowerbound = ALPHABETA_MINVAL;



	// This is the trick: make repeated calls to alphabeta, zeroing in on the
	// actual minimax value of theBoard by narrowing the bounds
	do {
		if ( currentEstimate == lowerbound ) {
			beta = currentEstimate + 1;
		} else
			beta = currentEstimate;

		UnrolledAlphabeta( theBoard, depth, beta - 1, beta, &Mov );
		currentEstimate = Mov.MoveEvaluation;
		
		if ( currentEstimate < beta )
			upperbound = currentEstimate;
			else
			lowerbound = currentEstimate;

    } while ( lowerbound < upperbound );
	
	bestMove->Set( &Mov );
	return bestMove->MoveEvaluation;
}




/**
 * The standard alphabeta, with the top level "unrolled" so that it can
 * return a Move structure instead of a mere minimax value
 * See AISearchAgent.Alphabeta for detailed comments on this code
*/
void AISearchAgentMTDF::UnrolledAlphabeta( Board *theBoard, int depth, int alpha, int beta, Move* bestMove )
{
	Move myBestMove;
	Move Bestmov;

	MoveListGenerator movegen;
	movegen.ComputeLegalMoves( theBoard );
	CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	Board newBoard;
	int bestSoFar;

	bestSoFar = ALPHABETA_MINVAL;
	int currentAlpha = alpha;
	Move *mov;

	// Loop on the successors
	while( ( mov = movegen.Next() ) != NULL ) {
		// Compute a board position resulting from the current successor
		newBoard.Clone( theBoard );
		newBoard.ApplyMove( mov );

		// And search it in turn
		int movScore;
		movScore = AlphaBeta( MINNODE, &newBoard, depth - 1, currentAlpha, beta );

		// Ignore illegal moves in the alphabeta evaluation
		if ( movScore == ALPHABETA_ILLEGAL ) {
			continue;
		}

		currentAlpha = max( currentAlpha, movScore );

		// Is the current successor better than the previous best?
		if ( movScore > bestSoFar ) {
			Bestmov.Set( mov );
			myBestMove.Set( mov );
			bestSoFar = movScore;
			myBestMove.MoveEvaluation = bestSoFar;

			// Can we cutoff now?
			if ( bestSoFar >= beta ) {

				TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_LOWERBOUND, depth, MoveCounter, &Bestmov, NumRegularNodes );

				// Add this move's efficiency in the HistoryTable
				CutoffHist.AddCount( theBoard->GetCurrentPlayer(), mov );
				bestMove->Set( &myBestMove );

				return;
			}
		}	
	}
	
	// Test for checkmate or stalemate
	if ( bestSoFar <= ALPHABETA_GIVEUP ) {
		newBoard.Clone( theBoard );
		MoveListGenerator secondary;
		newBoard.SwitchSides();
		if ( secondary.ComputeLegalMoves( &newBoard ) ) {
			// Then, we are not in check and may continue our efforts.
			CutoffHist.SortMoveList( &movegen, newBoard.GetCurrentPlayer() );
			movegen.ResetIterator();
			myBestMove.MoveType = Move::MOVE_STALEMATE;
			myBestMove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			while( ( mov = movegen.Next() ) != NULL ) {
				newBoard.Clone( theBoard );
				newBoard.ApplyMove( mov );
				if ( secondary.ComputeLegalMoves( &newBoard ) ) {
					myBestMove.MoveType = Move::MOVE_RESIGN;
				}
			}
		} else {
			// We're in check and our best hope is GIVEUP or worse, so either we are
			// already checkmated or will be soon, without hope of escape
			myBestMove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			myBestMove.MoveType = Move::MOVE_RESIGN;
		}
    }

	// If we haven't returned yet, we have found an accurate minimax score
	// for a position which is neither a checkmate nor a stalemate
	TransTables[ theBoard->CurrentPlayer ]->StoreBoard( theBoard, bestSoFar, Move::EVALTYPE_ACCURATE, depth, MoveCounter, &Bestmov, NumRegularNodes );

	bestMove->Set( &myBestMove );

	return;
}

void AISearchAgentMTDF::Print()
{
	*Output::cout << "MaxSearchNodes =     " << setw( 6 ) << MaxSearchSize << " nodes" << endl;
	*Output::cout << "MaxSearchTime =      " << setw( 6 ) << MaxSearchTime << " ms" << endl;
	*Output::cout << "MaxSearchIterTime =  " << setw( 6 ) << MaxSearchIterTime << " ms" << endl;
	*Output::cout << "MaxSearchDepth =     " << setw( 6 ) << MaxSearchDepth << " moves" << endl;
	*Output::cout << "TranspositionTable = " << setw( 6 ) << TranspositionTable::TABLE_SIZE << " entries ( " << ( double )( TranspositionTable::TABLE_SIZE * sizeof( TranspositionEntry ) ) / 1024.0 / 1024.0 << " MB )" << endl;
}
