//
// AISearchAgentMTDF - A sophisticated search agent
//
// Purpose: A ( mostly ) state-of-the-art search agent, implementing advanced
// techniques like the iterative-deepening MTDF search algorithm, transposition
// table, opening book and history table.
//
// History:
// 05.10.00 Completed initial version
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_AISEARCHAGENTMTDF_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_ )
#define AFX_AISEARCHAGENTMTDF_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "AISearchAgent.h"
#include "OpeningBook.h"

class AISearchAgentMTDF : public AISearchAgent  
{
private:
	// A reference to the game's opening book
	OpeningBook *Openings;

	/**
	 * Use the MTDF algorithm to find a good move.  MTDF repeatedly calls
	 * alphabeta with a zero-width search window, which creates very many quick
	 * cutoffs.  If alphabeta fails low, the next call will place the search
	 * window lower; in a sense, MTDF is a sort of binary search mechanism into
	 * the minimax space.
	 *	Return Value:
	 *			target
	 */
	int MTDF( Board *theBoard, int target, int depth, Move* bestMove );
	/**
	 * The standard alphabeta, with the top level "unrolled" so that it can
	 * return a Move structure instead of a mere minimax value
	 * See AISearchAgent.Alphabeta for detailed comments on this code
	 */
	void UnrolledAlphabeta( Board *theBoard, int depth, int alpha, int beta, Move* bestMove  );


public:
	static void Print();
	// A measure of the effort we are willing to expend on search
	static const int MaxSearchSize;
	static const int MaxSearchTime;
	static const int MaxSearchIterTime;
	static const int MaxSearchDepth;

	AISearchAgentMTDF( OpeningBook* ref );
	virtual ~AISearchAgentMTDF();
	virtual void PickBestMove( Board* theBoard, Move* bestMove ) ;
};

#endif // !defined( AFX_AISEARCHAGENTMTDF_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_ )
