// AISearchAgentNegaScout.cpp: Implementierung der Klasse AISearchAgentNegaScout.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "AISearchAgentNegaScout.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

const int AISearchAgentNegaScout::MinSearchDepth	=       5;
const int AISearchAgentNegaScout::MaxSearchDepth	=      25;
const int AISearchAgentNegaScout::MaxSearchSize		= 1100000;
const int AISearchAgentNegaScout::MaxSearchTime		=   11000;
const int AISearchAgentNegaScout::MaxSearchIterTime	=    3500;

// set verbosity to higher level to get debug information in logfile
int AISearchAgentNegaScout::verbosity = 0;

ofstream AISearchAgentNegaScout::logfile("NegaScout.log");

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/

AISearchAgentNegaScout::AISearchAgentNegaScout( OpeningBook* ref ) : AISearchAgent()
{
    Openings = ref;
}

AISearchAgentNegaScout::~AISearchAgentNegaScout()
{

}

/***************************************************************************
 * Operations
 **************************************************************************/

void AISearchAgentNegaScout::PickBestMove( Board* theBoard, Move* bestMove )
{
	MoveCounter++;
	Move* Mov;
	Timer theTimer;
	Timer iterTimer;


	// Compute efficiency statistics
	NumRegularNodes          = 0; NumQuiescenceNodes          = 0;
	NumRegularTTHits         = 0; NumQuiescenceTTHits         = 0;
	NumRegularCutoffs        = 0; NumQuiescenceCutoffs        = 0;
	NumRegularHashCollisions = 0; NumQuiescenceHashCollisions = 0;
	NumRegularMoves          = 0; NumQuiescenceMoves          = 0;
	NumUsedRegularMoves      = 0; NumUsedQuiescenceMoves      = 0;
	

	// First things first: look in the Opening Book, and if it contains a
	// move for this position, don't search anything
	Mov = Openings->Query( theBoard );
	if ( Mov != NULL ) {
		bestMove->Set( Mov );
		return;
	}

	// Store the identity of the moving side, so that we can tell Evaluator
	// from whose perspective we need to evaluate positions
	FromWhosePerspective = theBoard->GetCurrentPlayer();

	// clear CutoffHistory & HistoryTables
	CutoffHist.Forget();
	TransTable.SafeClear();
	TransTableBlack.SafeClear();

	// Begin search.  The search's maximum depth is determined on the fly,
	// according to how much effort has been spent; if it's possible to search
	// to depth 8 in 5 seconds, then by all means, do it!
	int bestGuess = 0;
	int iterdepth = 1;
	int NumAllNodes = 0;

	theTimer.Start();

	while( true ) {
		// Searching to depth 1 is not very effective, so we begin at 2
		iterdepth++;

		// Compute efficiency statistics
		NumRegularNodes          = 0; NumQuiescenceNodes          = 0;
		NumRegularTTHits         = 0; NumQuiescenceTTHits         = 0;
		NumRegularCutoffs        = 0; NumQuiescenceCutoffs        = 0;
		NumRegularHashCollisions = 0; NumQuiescenceHashCollisions = 0;
		NumRegularMoves          = 0; NumQuiescenceMoves          = 0;
		NumUsedRegularMoves      = 0; NumUsedQuiescenceMoves      = 0;

		// Look for a move at the current depth

		iterTimer.Start();
		NegaScoutDriver( theBoard, bestGuess, iterdepth, bestMove );
		iterTimer.Stop();
		
		// store the number of all nodes searched in this method call
		NumAllNodes += NumRegularNodes + NumQuiescenceNodes;
   
		bestGuess = bestMove->MoveEvaluation;

		// Feedback!
		*Output::cout <<  "Iteration of depth " << iterdepth << "; best move = ";
		bestMove->Print();
		// verbose 
		// *Output::cout << endl << BestMoves::Get( iterdepth, theBoard, &TransTable, &TransTableBlack ) << endl;
		
		// enable to show thinking ai
		// *Output::cout <<  " ^- AI-movesequence: " << BestMoves::Get(iterdepth, theBoard, &TransTable, &TransTableBlack) << " ( Score [white]: " << bestMove->MoveEvaluation / 100.0 << " )" << endl;

		
		// talk to Winboard if its active
		if ( WinBoardConnection::isActive ) {
			stringstream iostream;
			string smove;
			string sconn;
			string source;
			string dest;
			string sstream;
			string out;

			iostream.clear();
			iostream << iterdepth;
			iostream >> sstream;
			out = sstream + "\t";

			iostream.clear();
			iostream << bestGuess;
			iostream >> sstream;
			out = out + sstream + "\t0\t";

			iostream.clear();
			iostream << ( NumRegularNodes + NumQuiescenceNodes );
			iostream >> sstream;
			out = out + sstream + "\t";

			out = out + BestMoves::Get( iterdepth, theBoard, &TransTable, &TransTableBlack );
					
			WinBoardConnection::getInstance()->send( WinBoardConnection::CMD_THINKING, out );
		
			ShowStats();
			*Output::cout << "AI thought " << theTimer.GetDuration() / 1000 << " seconds! ( " << (NumAllNodes / theTimer.GetDuration()) << " KNodes / sec )" << endl;

			*Output::cout << " ---------------------------------" << endl << endl;

		}
		
		if ( ( iterdepth < MinSearchDepth ) ) continue;

		// Get out if we have searched deep enough
		if ( ( NumRegularNodes + NumQuiescenceNodes ) > MaxSearchSize )
			break;
		if ( iterdepth >= MaxSearchDepth )
			break;
		if ( iterTimer.GetDuration() > MaxSearchIterTime )
			break;
		if ( theTimer.GetDuration() > MaxSearchTime )
			break;
		if ( bestMove->MoveEvaluation >= AISearchAgent::ALPHABETA_MAXVAL - 100 )
			break;
	}

	theTimer.Stop();

/*
	stringstream iostream;
	string filename;
	iostream << theBoard->MoveNo;
	iostream >> filename;

	filename = "TT-white " + filename + ".txt";

	TransTable.PrintToFile(filename.c_str());

	iostream.clear();
	iostream << theBoard->MoveNo;
	iostream >> filename;

	filename = "TT-black " + filename + ".txt";

	TransTableBlack.PrintToFile(filename.c_str());
*/
	*Output::cout << "AI thought " << theTimer.GetDuration() / 1000 << " seconds! ( " << (NumAllNodes / theTimer.GetDuration()) << " KNodes / sec )" << endl;

	return;
}



// simple method which encapsulte the search 
int AISearchAgentNegaScout::NegaScoutDriver( Board *theBoard, int target, int depth, Move* bestMove )
{
	if (verbosity > 0) {
		logfile << endl << "ITERATION FOR DEPTH " << depth << endl;
	}
	
	int windowwidth = 40;
	int best;
	// search at first at a small window -> is much more faster that a complete search
	int alpha = target - windowwidth;
	int beta =  target + windowwidth;
	best = UnrolledNegaScout(theBoard, depth, bestMove, alpha, beta);
	
	// if quick search fails (return value not in lower and upper bound) -> research with full window
	if ( best <= alpha || best >= beta)
		best = UnrolledNegaScout(theBoard, depth, bestMove, ALPHABETA_MINVAL, ALPHABETA_MAXVAL);
	
	bestMove->MoveEvaluation = best;
	
	if (verbosity > 0) {
		logfile << "Score: " << best << "\tNumNodes: " << NumRegularNodes << "\tNumCutoffs: " << NumRegularCutoffs << endl;
	}

	return best;
}



/**
 * The standard NegaScout, with the top level "unrolled" so that it can
 * return a Move structure instead of a mere minimax value
*/
int AISearchAgentNegaScout::UnrolledNegaScout( Board *theBoard, int depth, Move *move, int alpha, int beta ) {
	
	Move bestmove;
	NumRegularNodes++;
	int best = ALPHABETA_MINVAL - 1;
	TranspositionEntry *theEntry = NULL;
	bool allillegal = true;
	
	// compute the haslock key and store it - it won't change since theBoard isn't changed
	int hashlock  = theBoard->HashLock();
	
	// get the TransositionEntry for theBoard
	// two different TranspositionTables are used: one for white - one for black
	TransTables[ theBoard->CurrentPlayer ]->LookupBoard( theBoard, &theEntry );

	int tempalpha = alpha;
	int tempbeta  = beta;
	int current;
	int bestwhen = 0;
	int tempcurrent;
	int moveindex = 0;
	int noderesult = ALPHABETA_NOCUT;
	Move *tempmove;
	Board newBoard;
	MoveListGenerator movegen;
	Move subMove;


    // generate successors and search them in turn
    // If ComputeLegalMoves returns false, then the current position is illegal
    // because one or more moves could capture a king!    

    if ( !movegen.ComputeLegalMoves( theBoard ) ) {

		if ( verbosity > 10 ) {
			logfile << "illegal position" << endl; 
		}
		
		return ALPHABETA_ILLEGAL;
	}

	// sort moves to get an early cutoff (<- hopefully)
	CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );
	if ( verbosity > 25 ) {
		movegen.Print();
	}

	// Loop on the successors
	while ( ( ( tempmove = movegen.Next() ) != NULL ) && ( best < beta ) && ( best < (ALPHABETA_MAXVAL - 100)) ) {
		// Compute a board position resulting from the current successor
		newBoard.Clone( theBoard );
		newBoard.ApplyMove( tempmove );

		// required for statistics
		moveindex++;

		if ( depth > 1 ) {
			current =
				-NegaScout( &newBoard, depth - 1, &subMove, -tempbeta, -tempalpha );
			if ( current != ALPHABETA_ILLEGAL )
				allillegal = false;
		} else { // depth == 1
			// If we have reached the maximum depth of the search, stop recursion
			// and begin quiescence search
			current = -QuiescentSearch( &newBoard, -tempbeta, -tempalpha, 0);
			if ( current != ALPHABETA_ILLEGAL ) {
				allillegal = false;
			}

		}
		if ( ( current > tempalpha ) && ( current < beta ) ) {
			if ( depth > 1 ) {
				tempcurrent = 
					-NegaScout( &newBoard, depth - 1, &subMove, -beta, -current);
			} else { // depth == 1
				// If we have reached the maximum depth of the search, stop recursion
				// and begin quiescence search
				tempcurrent = -QuiescentSearch( &newBoard, -beta, -current, 0);
			}
			if ( tempcurrent > current ) current = tempcurrent;
		}

		if ( current > best ) {
			best = current;
			bestwhen = moveindex;
			bestmove.Set(tempmove);
			if ( best > tempalpha ) {
				tempalpha = best;
			}

		}
		tempbeta = tempalpha + 1;
	}

	// Test for checkmate or stalemate
	if ( best <= ALPHABETA_GIVEUP ) {
		newBoard.Clone( theBoard );
		MoveListGenerator secondary;
		newBoard.SwitchSides();
		if ( secondary.ComputeLegalMoves( &newBoard ) ) {
			// Then, we are not in check and may continue our efforts.
			CutoffHist.SortMoveList( &movegen, newBoard.GetCurrentPlayer() );
			movegen.ResetIterator();
			bestmove.MoveType = Move::MOVE_STALEMATE;
			bestmove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			while( ( tempmove = movegen.Next() ) != NULL ) {
				newBoard.Clone( theBoard );
				newBoard.ApplyMove( tempmove );
				if ( secondary.ComputeLegalMoves( &newBoard ) ) {
					bestmove.MoveType = Move::MOVE_RESIGN;
				}
			}
		} else {
			// We're in check and our best hope is GIVEUP or worse, so either we are
			// already checkmated or will be soon, without hope of escape
			bestmove.MovingPiece = Board::KING + theBoard->GetCurrentPlayer();
			bestmove.MoveType = Move::MOVE_RESIGN;
		}
    }

	move->Set(&bestmove);

	// what kind of result did we get ?
	if ( best >= beta ) { // cutoff ?
		noderesult = ALPHABETA_BETACUT;
		NumRegularCutoffs++;
		CutoffHist.AddCount( theBoard->CurrentPlayer, move, depth );
	} else {
		if ( best <= alpha ) {
			noderesult = ALPHABETA_BELOWALPHA;
		} else {
			// if ( ( best > alpha ) && ( best < alpha ) )
			noderesult = ALPHABETA_NOCUT;
		}
	}

	// store to TranspositionTable if the move found is more important that the old one stored there
	// theEntry is still valid because theBoard isn't changed
	if ( ( theEntry->timeStamp + theEntry->theDepth < theBoard->MoveNo + depth )
		|| ( ( theEntry->theDepth < depth ) && ( theEntry->theLock == hashlock ) )
		|| ( ( noderesult == ALPHABETA_NOCUT ) && ( theEntry->theEvalType != Move::EVALTYPE_ACCURATE ) )
		|| ( theEntry->theEvalType == TranspositionEntry::NULL_ENTRY ) 
		|| ( theEntry->theDepth == 0 ) ) {
		if ( noderesult == ALPHABETA_BETACUT ) {
			theEntry->theEvalType = Move::EVALTYPE_LOWERBOUND;
		} else {
			if ( noderesult == ALPHABETA_BELOWALPHA ) {
				theEntry->theEvalType = Move::EVALTYPE_UPPERBOUND;
			} else {
				// if ( ( best > alpha ) && ( best < beta ) )
				theEntry->theEvalType = Move::EVALTYPE_ACCURATE;
			}
		}
		
		// update entries...
		theEntry->theEval	= best;
		theEntry->timeStamp = theBoard->MoveNo;
		theEntry->theLock	= hashlock;
		theEntry->theDepth	= depth;
		theEntry->theSourceSquare      = bestmove.SourceSquare;
		theEntry->theDestinationSquare = bestmove.DestinationSquare;
		theEntry->theNode	= NumRegularNodes;
	}

	NumRegularMoves += movegen.Size() - 1;
	NumUsedRegularMoves += bestwhen - 1;

	if ( ( verbosity > 8 ) && ( depth > 3 ) ) {
		logfile << "checked " << bestwhen << " of " << movegen.Size() << " at depth "<< depth << endl;
	}

	move->Set(&bestmove);

	return best;
}

int AISearchAgentNegaScout::NegaScout( Board *theBoard, int depth, Move *move, int alpha, int beta )
{

	bool player_max;
	string PlayerArray[2];
	PlayerArray[0] =  " MAX";
	PlayerArray[1] =  " MIN";
	if ( theBoard->CurrentPlayer == FromWhosePerspective ) player_max = 0; else player_max = 1;
	int SuggestedSourceSquare      = Move::NULL_MOVE;
	int SuggestedDestinationSquare = Move::NULL_MOVE;

	Move bestmove;
	Move* SuggestedMove = NULL;
    
	// Count the number of nodes visited in the full-width search
	NumRegularNodes++;
	
	int best = ALPHABETA_MINVAL - 1;
	TranspositionEntry *theEntry = NULL;
	bool allillegal = true;
	int hashlock  = theBoard->HashLock();
    // First things first: let's see if there is already something useful
    // in the transposition table, which might save us from having to search
    // anything at all
	if ( ( TransTables[ theBoard->CurrentPlayer ]->LookupBoard( theBoard, &theEntry ) ) && ( theEntry->theDepth >= depth ) ) {
		NumRegularTTHits++;
		if ( theEntry->theEvalType == Move::EVALTYPE_ACCURATE ) {
			if ( verbosity > 15 ) {
				logfile << "hashed exact level " << theEntry->theDepth << "\t\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			move->SourceSquare = theEntry->theSourceSquare;
			move->DestinationSquare = theEntry->theDestinationSquare;

			NumRegularCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				return theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			}
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				return theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			}
			return theEntry->theEval; // = theLowerBound == theUpperbound;  
		}
		if ( ( theEntry->theEval >= beta  ) && ( theEntry->theEvalType == Move::EVALTYPE_LOWERBOUND ) ) {
			if ( verbosity > 15 ) {
				logfile << "hashed beta cut off level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			move->SourceSquare = theEntry->theSourceSquare;
			move->DestinationSquare = theEntry->theDestinationSquare;

			NumRegularCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				return theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			}
			return theEntry->theEval; // = theLowerBound;
		}
		if ( ( theEntry->theEval <= alpha ) && ( theEntry->theEvalType == Move::EVALTYPE_UPPERBOUND ) ) {
			if ( verbosity > 15 ) {
				logfile << "hashed alpha cut off level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			move->SourceSquare = theEntry->theSourceSquare;
			move->DestinationSquare = theEntry->theDestinationSquare;

			NumRegularCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				return theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			}
			return theEntry->theEval; // = theUpperBound;
		}
		if ( ( theEntry->theEvalType == Move::EVALTYPE_LOWERBOUND) && ( theEntry->theEval > alpha ) ) {
			if ( verbosity > 15 ) {
				logfile << "modified alpha by hash level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				alpha =  theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			} else
				alpha = theEntry->theEval; // = theLowerBound;
			bestmove.SourceSquare = theEntry->theSourceSquare;
			bestmove.DestinationSquare = theEntry->theDestinationSquare;

			best = alpha;
            SuggestedSourceSquare      = theEntry->theSourceSquare;
			SuggestedDestinationSquare = theEntry->theDestinationSquare;
		}
		if ( ( theEntry->theEvalType == Move::EVALTYPE_UPPERBOUND) && ( theEntry->theEval < beta ) ) {
			if ( verbosity > 15 ) {
				logfile << "modified beta by hash level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				beta = theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			} else
				beta = theEntry->theEval; // = theUpperBound;
            SuggestedSourceSquare      = theEntry->theSourceSquare;
			SuggestedDestinationSquare = theEntry->theDestinationSquare;
		}
	}


	if ( ( ( theEntry->theEvalType != TranspositionEntry::NULL_ENTRY ) && ( hashlock != theEntry->theLock ) ) ) {
		NumRegularHashCollisions++;
		if ( verbosity > 15 )
			logfile << "hash collision\thash key " << hashlock << endl; 
	}

	int tempalpha = alpha;
	int tempbeta  = beta;
	int current;
	int bestwhen = 0;
	int tempcurrent;
	int moveindex = 0;
	int noderesult = ALPHABETA_NOCUT;
	Move *tempmove;
	Board newBoard;
	MoveListGenerator movegen;
	Move subMove;

    // generate successors and search them in turn
    // If ComputeLegalMoves returns false, then the current position is illegal
    // because one or more moves could capture a king!    
	if ( !movegen.ComputeLegalMoves( theBoard ) ) {

		if ( verbosity > 10 ) {
			logfile << "illegal position" << endl; 
		}
		
		// -ALPHABETA_ILLEGAL because the sign is switched by  
		// method call of -NegaScout(...)
		return -ALPHABETA_ILLEGAL;
	}

	CutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	if ( SuggestedSourceSquare != Move::NULL_MOVE ) {
		if ( ( SuggestedMove = movegen.FindMoveForSquares( SuggestedSourceSquare, SuggestedDestinationSquare ) ) == NULL ) {
			if ( verbosity > 10 ) {
				logfile << "wrong TT Entry !!!" << endl; 
			}
			movegen.ResetIterator();
			tempmove = movegen.Next();
		} else {
			movegen.ResetIterator();
			tempmove = SuggestedMove;
		}
	} else {
		movegen.ResetIterator();
		tempmove = movegen.Next();
	}

	if ( verbosity > 25 ) {
		movegen.Print();
	}

	// Loop on the successors
	while ( ( tempmove != NULL ) && ( best < beta ) ) {
		// Compute a board position resulting from the current successor
		newBoard.Clone( theBoard );
		newBoard.ApplyMove( tempmove );

		// required for statistics
		moveindex++;

		if ( depth > 1 ) {
			current =
				-NegaScout( &newBoard, depth - 1, &subMove, -tempbeta, -tempalpha );
			if ( current != ALPHABETA_ILLEGAL )
				allillegal = false;
		} else { // depth == 1
			// If we have reached the maximum depth of the search, stop recursion
			// and begin quiescence search

			current = -QuiescentSearch( &newBoard, -tempbeta, -tempalpha, 0);
			if ( current != ALPHABETA_ILLEGAL ) {
				allillegal = false;
			}
		}
		if ( ( current > tempalpha ) && ( current < beta ) ) {
			if ( depth > 1 ) {
				tempcurrent = 
					-NegaScout( &newBoard, depth - 1, &subMove, -beta, -current);
			} else { // depth == 1
				// If we have reached the maximum depth of the search, stop recursion
			    // and begin quiescence search

				tempcurrent = -QuiescentSearch( &newBoard, -beta, -current, 0);
			}
			if ( tempcurrent > current ) current = tempcurrent;
		}

		if ( current > best ) {
			best = current;
			bestwhen = moveindex;
			bestmove.SourceSquare = tempmove->SourceSquare;
			bestmove.DestinationSquare = tempmove->DestinationSquare;

			if ( best > tempalpha ) {
				tempalpha = best;
			}
		}
		tempbeta = tempalpha + 1;

		tempmove = movegen.Next();
		if ( ( tempmove == SuggestedMove ) && ( tempmove != NULL ) ) tempmove = movegen.Next();
	}

	// Test for checkmate or stalemate
	// Both cases occur if and only if there is no legal move for MAX, i.e.,
	// if "bestSoFar" is ALPHABETA_MINVAL.  There are two cases: we
	// have checkmate ( in which case the score is accurate ) or stalemate ( in
	// which case the position should be re-scored as a draw with value 0.
	if ( best <= ALPHABETA_MINVAL ) {

		// Can MIN capture MAX's king?  First, ask the machine to generate
		// moves for MIN
		newBoard.Clone( theBoard );
		newBoard.SwitchSides();

		// And if one of MIN's moves is a king capture, indicating that the
		// position is illegal, we have checkmate and must return MINVAL.  We
		// add the depth simply to "favor" delaying tactics: a mate in 5 will
		// score higher than a mate in 3, because the likelihood that the
		// opponent will miss it is higher; might as well make life difficult!
		if ( !movegen.ComputeLegalMoves( &newBoard ) ) {
			return best + depth;
		} else {
			return 0;
		}
	}	

	move->Set(&bestmove);
	
	// what kind of result did we get ?
	if ( best >= beta ) { // cutoff ?
		noderesult = ALPHABETA_BETACUT;
		NumRegularCutoffs++;
		// Add this move's efficiency in the CutoffTable
		CutoffHist.AddCount( theBoard->CurrentPlayer, move, depth );
	} else {
		if ( best <= alpha ) {
			noderesult = ALPHABETA_BELOWALPHA;
		} else {
			// if ( ( best > alpha ) && ( best < alpha ) )
			noderesult = ALPHABETA_NOCUT;
		}
	}

	// store to TranspositionTable
	if ( ( theEntry->timeStamp + theEntry->theDepth < theBoard->MoveNo + depth )
		|| ( ( theEntry->theDepth < depth ) && ( theEntry->theLock == hashlock ) )
		|| ( ( noderesult == ALPHABETA_NOCUT ) && ( theEntry->theEvalType != Move::EVALTYPE_ACCURATE ) )
		|| ( theEntry->theEvalType == TranspositionEntry::NULL_ENTRY ) ) {
		if ( noderesult == ALPHABETA_BETACUT ) {
			theEntry->theEvalType = Move::EVALTYPE_LOWERBOUND;
		} else {
			if ( noderesult == ALPHABETA_BELOWALPHA ) {
				theEntry->theEvalType = Move::EVALTYPE_UPPERBOUND;
			} else {
				// if ( ( best > alpha ) && ( best < beta ) )
				theEntry->theEvalType = Move::EVALTYPE_ACCURATE;
			}
		}

		theEntry->theEval	= best;
		theEntry->timeStamp = theBoard->MoveNo;
		theEntry->theLock	= hashlock;
		theEntry->theDepth	= depth;
		theEntry->theSourceSquare		= bestmove.SourceSquare;
		theEntry->theDestinationSquare	= bestmove.DestinationSquare;
		theEntry->theNode	= NumRegularNodes;
	}

	NumRegularMoves     += movegen.Size() - 1;
	NumUsedRegularMoves += bestwhen - 1;
	
	if ( ( verbosity > 8 ) && ( depth > 3 ) ) {
		logfile << "checked " << bestwhen << " of " << movegen.Size() << " at depth "<< depth << endl;
	} 
	return best;
}

/**
 * A slight variant of NegaScout which only considers captures and null moves
 * This is necesary because the evaluation function can only be applied to
 * "quiet" positions where the tactical situation ( i.e., material balance ) is
 * unlikely to change in the near future.
 * Note that, in this version of the code, the quiescence search is not limited
 * by depth; we continue digging for as long as we can find captures.  Some other
 * programs impose a depth limit for time-management purposes.
 */
int AISearchAgentNegaScout::QuiescentSearch( Board *theBoard, int alpha, int beta, int depth )
{
	bool player_max;
	string PlayerArray[2];
	PlayerArray[0] =  " MAX";
	PlayerArray[1] =  " MIN";
	if ( theBoard->CurrentPlayer == FromWhosePerspective ) player_max = 0; else player_max = 1;
	int SuggestedSourceSquare      = Move::NULL_MOVE;
	int SuggestedDestinationSquare = Move::NULL_MOVE;
	Move *SuggestedMove = NULL;

	Move bestmove;
	NumQuiescenceNodes++;
	int best = ALPHABETA_MINVAL - 1;
	TranspositionEntry *theEntry = NULL;
	bool allillegal = true;
	int hashlock  = theBoard->HashLock();

    // First things first: let's see if there is already something useful
    // in the transposition table, which might save us from having to search
    // anything at all
	if ( ( TransTables[ theBoard->CurrentPlayer ]->LookupBoard( theBoard, &theEntry ) ) && ( theEntry->theDepth >= 0 ) ) {
		NumQuiescenceTTHits++;
		if ( theEntry->theEvalType == Move::EVALTYPE_ACCURATE ) {
			if ( verbosity > 15 ) {
				logfile << "qhashed exact level " << theEntry->theDepth << "\t\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			NumQuiescenceCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				return theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			}
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				return theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			}
			return theEntry->theEval;  // = theLowerBound == theUpperBound;  
		}
		if ( ( theEntry->theEval >= beta  ) && ( theEntry->theEvalType == Move::EVALTYPE_LOWERBOUND ) ) {
			if ( verbosity > 15 ) {
				logfile << "qhashed beta cut off level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			NumQuiescenceCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				return theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			}
			return theEntry->theEval;  // = theLowerBound;
		}
		if ( ( theEntry->theEval <= alpha ) && ( theEntry->theEvalType == Move::EVALTYPE_UPPERBOUND ) ) {
			if ( verbosity > 15 ) {
				logfile << "qhashed alpha cut off level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
			NumQuiescenceCutoffs++;
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				return theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			}
			return theEntry->theEval;  // = theUpperBound;
		}
		if ( ( theEntry->theEvalType == Move::EVALTYPE_LOWERBOUND) && ( theEntry->theEval > alpha ) ) {
			if ( verbosity > 15 ) {
				logfile << "qmodified alpha by hash level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
            SuggestedSourceSquare      = theEntry->theSourceSquare;
			SuggestedDestinationSquare = theEntry->theDestinationSquare;
			if (theEntry->theEval >= ALPHABETA_MAXVAL - 50) {
				alpha = theEntry->theEval + theEntry->timeStamp - theBoard->MoveNo;
			} else
				alpha = theEntry->theEval; // = theLowerBound;
		}
		if ( ( theEntry->theEvalType == Move::EVALTYPE_UPPERBOUND) && ( theEntry->theEval < beta ) ) {
			if ( verbosity > 15 ) {
				logfile << "qmodified beta by hash level " << theEntry->theDepth << "\t value " << theEntry->theEval << "\t hash key " << theEntry->theLock << "\t" << PlayerArray[player_max] << " @ " << theEntry->theNode << endl;
			}
            SuggestedSourceSquare      = theEntry->theSourceSquare;
			SuggestedDestinationSquare = theEntry->theDestinationSquare;
			if (theEntry->theEval >= ALPHABETA_MINVAL + 50) {
				beta = theEntry->theEval - theEntry->timeStamp + theBoard->MoveNo;
			} else
				beta = theEntry->theEval;  // = theUpperBound;
		}
	}

	if ( ( ( theEntry->theEvalType != TranspositionEntry::NULL_ENTRY ) && ( hashlock != theEntry->theLock ) ) ) {
		NumQuiescenceHashCollisions++;
		if ( verbosity > 15 )
			logfile << "hash collision\thash key " << hashlock << endl; 
	}

	int tempalpha = alpha;
	int current;
	int bestwhen = 0;
	int moveindex = 0;
	int noderesult = ALPHABETA_NOCUT;
	Move *tempmove;
	Board newBoard;
	MoveListGenerator movegen;
	Move subMove;
	
    // Start with evaluation of the null-move, just to see whether it is more
    // effective than any capture, in which case we must stop looking at
    // captures and damaging our position
	current = Evaluator->EvaluateComplete( theBoard, FromWhosePerspective, false );
	if ( theBoard->CurrentPlayer != FromWhosePerspective ) current = -current;

    if ( !movegen.ComputeQuiescenceMoves( theBoard ) ) {
		// quiet position
		return current;
	}

	if ( ( current >= beta ) || ( movegen.Size() == 0 ) ) {
		return current;
	}

	if ( current > best ) {
		best = current;
		if ( best > tempalpha ) {
			tempalpha = best;
		}
	}

	// sort moves <<< here
	qCutoffHist.SortMoveList( &movegen, theBoard->GetCurrentPlayer() );

	if ( SuggestedSourceSquare != Move::NULL_MOVE ) {
		if ( ( SuggestedMove = movegen.FindMoveForSquares( SuggestedSourceSquare, SuggestedDestinationSquare ) ) == NULL ) {
			if ( verbosity > 10 ) {
				logfile << "wrong TT Entry !!!" << endl; 
			}
			movegen.ResetIterator();
			tempmove = movegen.Next();
		} else {
			movegen.ResetIterator();
			tempmove = SuggestedMove;
		}
	} else {
		movegen.ResetIterator();
		tempmove = movegen.Next();
	}

	// Loop on the successors
	while ( ( tempmove != NULL ) && ( best < beta ) ) {
		// Compute a board position resulting from the current successor
		newBoard.Clone( theBoard );
		newBoard.ApplyMove( tempmove );

		// required for statistics
		moveindex++;

		current = -QuiescentSearch( &newBoard, -beta, -tempalpha, depth - 1 );

		if ( current > best ) {
			best = current;
			bestwhen = moveindex;
			bestmove.SourceSquare = tempmove->SourceSquare;
			bestmove.DestinationSquare = tempmove->DestinationSquare;
			if ( best > tempalpha ) {
				tempalpha = best;
			}
		}

		tempmove = movegen.Next();
		if ( ( tempmove == SuggestedMove ) && ( tempmove != NULL ) ) tempmove = movegen.Next();
	}

	// Test for checkmate or stalemate
	// Both cases occur if and only if there is no legal move for MAX, i.e.,
	// if "bestSoFar" is ALPHABETA_MINVAL.  There are two cases: we
	// have checkmate ( in which case the score is accurate ) or stalemate ( in
	// which case the position should be re-scored as a draw with value 0.
	if ( best <= ALPHABETA_MINVAL ) {

		// Can MIN capture MAX's king?  First, ask the machine to generate
		// moves for MIN
		newBoard.Clone( theBoard );
		newBoard.SwitchSides();

		// And if one of MIN's moves is a king capture, indicating that the
		// position is illegal, we have checkmate and must return MINVAL.  We
		// add the depth simply to "favor" delaying tactics: a mate in 5 will
		// score higher than a mate in 3, because the likelihood that the
		// opponent will miss it is higher; might as well make life difficult!
		if ( !movegen.ComputeLegalMoves( &newBoard ) ) {
			return best + depth;
		} else {
			return 0;
		}
	}	

	int flags;
	

	// what kind of result did we get ?
	if ( best >= beta ) {
		NumQuiescenceCutoffs++;
		flags = ALPHABETA_LOWCUT;
		qCutoffHist.AddCount( theBoard->CurrentPlayer, &bestmove, 1 );
	} else {
		if ( best <= alpha ) {
			flags = ALPHABETA_HIGHCUT;
		} else {
			flags = ALPHABETA_NOCUT;
		}
	}

	// Store this best move in the TransTable - but only if the entry is empty
	if ( ( theEntry->theEvalType == TranspositionEntry::NULL_ENTRY ) ) {
		if ( flags == ALPHABETA_LOWCUT ) {
			theEntry->theEvalType = Move::EVALTYPE_LOWERBOUND;
		} else {
			if ( flags == ALPHABETA_HIGHCUT ) {
				theEntry->theEvalType = Move::EVALTYPE_UPPERBOUND;
			} else {
				theEntry->theEvalType = Move::EVALTYPE_ACCURATE;
			}
		}

		theEntry->theEval	= best;
		theEntry->timeStamp = theBoard->MoveNo;
		theEntry->theLock	= hashlock;
		theEntry->theDepth	= depth;
		theEntry->theSourceSquare		= bestmove.SourceSquare;
		theEntry->theDestinationSquare	= bestmove.DestinationSquare;
		theEntry->theNode	= NumRegularNodes;
	}

	if ( bestwhen != 0 ) {
		NumQuiescenceMoves += movegen.Size() - 1;
		NumUsedQuiescenceMoves += bestwhen - 1;
	}

	if ( verbosity > 13 ) {
		logfile << "QUIESCENT - checked " << bestwhen << " of " << movegen.Size() << endl;
	} 
	return best;
}

void AISearchAgentNegaScout::Print()
{
	*Output::cout << "MaxSearchNodes =     " << setw( 7 ) << MaxSearchSize		<< " nodes" << endl;
	*Output::cout << "MaxSearchTime =      " << setw( 7 ) << MaxSearchTime		<< " ms" << endl;
	*Output::cout << "MaxSearchIterTime =  " << setw( 7 ) << MaxSearchIterTime	<< " ms" << endl;
	*Output::cout << "MinSearchDepth =     " << setw( 7 ) << MinSearchDepth		<< " moves" << endl;
	*Output::cout << "MaxSearchDepth =     " << setw( 7 ) << MaxSearchDepth		<< " moves" << endl;
	*Output::cout << "TranspositionTable = " << setw( 7 ) << TranspositionTable::TABLE_SIZE << " x2 entries ( " << ( double )( TranspositionTable::TABLE_SIZE * 2 * sizeof( TranspositionEntry ) ) / 1024.0 / 1024.0 << " MB )" << endl;
}

istream& AISearchAgentNegaScout::read( istream& is )
{
	AISearchAgent::read(is);
	TransTableBlack.Clear();
	// enable the following lines are debugging options
	// TransTableBlack.read( is );
	return is;
}

ostream& AISearchAgentNegaScout::write( ostream& os ) const
{
	AISearchAgent::write(os);
	// enable the following lines are debugging options
	// TransTableBlack.write( os );
	
	return os;
}

void AISearchAgentNegaScout::ShowStats()
{
	*Output::cout << endl << " S T A T S  -  NegaScout" << endl;
	*Output::cout << "Regular nodes:   " << setw( 7 ) << NumRegularNodes		  << "  Quiescence nodes:   " << setw( 7 ) << NumQuiescenceNodes		  << endl;
	*Output::cout << "Regular TTHits:  " << setw( 7 ) << NumRegularTTHits		  << "  Quiescence TTHits:  " << setw( 7 ) << NumQuiescenceTTHits		  << endl;
	*Output::cout << "Regular Cutoffs: " << setw( 7 ) << NumRegularCutoffs		  << "  Quiescence Cutoffs: " << setw( 7 ) << NumQuiescenceCutoffs		  << endl;
	*Output::cout << "Regular Hash col:" << setw( 7 ) << NumRegularHashCollisions << "  Quiescence Hash col:" << setw( 7 ) << NumQuiescenceHashCollisions << endl;
	if ( NumRegularMoves != 0 && NumQuiescenceMoves != 0 )
		*Output::cout << "Regular Moveperf:" << setw( 7 ) << ( 100.00 - ( NumUsedRegularMoves * 10000 / NumRegularMoves / 100.0 ) ) 
																					  << "% Quiescence Moveperf:" << setw( 7 ) << (100.0 - (NumUsedQuiescenceMoves * 10000 / NumQuiescenceMoves / 100.0 )) << "%" << endl;
	TransTable.ShowStats();
	TransTableBlack.ShowStats();

}
