//
// AISearchAgentNegaScout - A sophisticated search agent
//
// Purpose: A ( mostly ) state-of-the-art search agent, implementing advanced
// techniques like the iterative-deepening MTDF search algorithm, transposition
// table, opening book and history table.
//
// History:
// 05.10.00 Completed initial version
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_AISearchAgentNegaScout_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_ )
#define AFX_AISearchAgentNegaScout_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "AISearchAgent.h"
#include "OpeningBook.h"

class AISearchAgentNegaScout : public AISearchAgent  
{

public:
	// A measure of the effort we are willing to expend on search
	static const int MaxSearchSize;
	static const int MaxSearchTime;
	static const int MaxSearchIterTime;
	static const int MinSearchDepth;
	static const int MaxSearchDepth;
	static int verbosity;
	static ofstream logfile;


private:
	// A reference to the game's opening book
	OpeningBook *Openings;

	/**
	 * Use the MTDF algorithm to find a good move.  MTDF repeatedly calls
	 * alphabeta with a zero-width search window, which creates very many quick
	 * cutoffs.  If alphabeta fails low, the next call will place the search
	 * window lower; in a sense, MTDF is a sort of binary search mechanism into
	 * the minimax space.
	 *	Return Value:
	 *			target
	 */
	/**
	 * The standard alphabeta, with the top level "unrolled" so that it can
	 * return a Move structure instead of a mere minimax value
	 * See AISearchAgent.Alphabeta for detailed comments on this code
	 */


/***********************************************************************/
/*                             Operationen                             */
/***********************************************************************/

public:
	static void Print();

	AISearchAgentNegaScout( OpeningBook* ref );
	virtual ~AISearchAgentNegaScout();

	int NegaScout(Board *theBoard, int depth, Move *bestMove, int alpha, int beta);
	
	int NegaScoutDriver( Board *theBoard, int target, int depth, Move* bestMove );
	
	istream& read( istream& is );
	
	virtual void PickBestMove( Board* theBoard, Move* bestMove ) ;
	
	int QuiescentSearch( Board *theBoard, int alpha, int beta, int depth );
	
	void ShowStats();
	
	int UnrolledNegaScout( Board *theBoard, int depth, Move *bestMove, int alpha, int beta );
	
	ostream& write( ostream& os ) const;
};

#endif // !defined( AFX_AISearchAgentNegaScout_H__918F5BA0_C996_4240_A6E3_1563627C47AC__INCLUDED_ )
