// BaChess.cpp : Definiert den Einsprungpunkt fr die Konsolenanwendung.
//

#include "stdafx.h"      
#include "Board.h"
#include "Player.h"

typedef std::vector<int> intvector;

/***************************************************************************
 * Global Variables
 **************************************************************************/
ofstream*	WinBoardConsole;

int handleMemoryDepletion( size_t size )
{
	cerr << "Running out of memory! Tried to allocate " << size / 1024 << " KB.";
	exit( -1 );
	return 0;
}

void init()
{
	_set_new_handler( handleMemoryDepletion );
	
	Output::cout = &cout;
	
	Board::InitStatics();
	MoveListGenerator::InitStatics();
}



int main( int argc, char* argv[] )
{
// initialize random number generator
	int randomSeed = GetTickCount();
 	srand( randomSeed );   

	init();

	*Output::cout << "            Welcome to Chess!" << endl;
	*Output::cout << "            Version: " << Game::Version << endl;
	*Output::cout << "			  id: " << randomSeed << endl;
	*Output::cout << "    ~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~~" << endl;

	/** Let's play a little bit! */

	// checking command line options
	if ( argc > 1 ) {
		if ( strcmp( argv[ 1 ], "xboard" ) == 0 ) {
			WinBoardConnection::isActive = true;
			new WinBoardConnection();
			WinBoardConsole = new ofstream( "BaChessConsole.txt" );
			Output::cout = WinBoardConsole;
		}

		if ( strcmp( argv[ 1 ], "createbook" ) == 0 ) {
			OpeningBook::StartBookCreation( argc, argv );
			return 0;
		}
	}

	// initialize new game
	// parameters: openingbook, output gamefile
	Game theGame ( "openingbook.bin", "games.pgn" );

	// continue checking command line options
	if ( argc > 1 ) {
		if ( strcmp( argv[ 1 ], "load" ) == 0 ) {
			theGame.SetupGameBoard( argv[ 2 ] );
		}
	}

	// create players for playing :- )
	if ( WinBoardConnection::isActive ) {
//		theGame.SetPlayerWhite ( Game::AIWINBOARD, "Whity" );
		theGame.SetPlayerWhite ( Game::HUMANWINBOARD, "Whity" );
//		theGame.SetPlayerBlack ( Game::HUMANWINBOARD, "Blacky" );
//		theGame.SetPlayerBlack ( Game::AIWINBOARD, "Blacky" );
	} else {
		theGame.SetPlayerWhite ( Game::HUMAN, "Whity" );
		theGame.SetPlayerBlack ( Game::AI, "Blacky" );
	}

	// show AI settings
	AISearchAgentNegaScout::Print();

	// start the game
	theGame.RunGame();

	/* clean up */
	if ( WinBoardConnection::isActive ) {
		delete WinBoardConnection::getInstance();
		Output::cout = &cout;
		WinBoardConsole->close();
		delete WinBoardConsole;
	}

	Board::DeleteStatics();
	MoveListGenerator::DeleteStatics();
	return 0;
}

