// BestMoves.cpp: Implementierung der Klasse BestMoves.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "BestMoves.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#endif

BestMoves** BestMoves::pthis = NULL;
const int BestMoves::MaxSearchDepth = AISearchAgentMTDF::MaxSearchDepth;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
/*
BestMoves::BestMoves()
{
}

BestMoves::~BestMoves()
{
}

*/

BestMoves::BestMoves( int depth )
{
	storedMoves = 0;
	this->depth = depth;
	Moves = new BestMoveEntry[ depth ];
}

BestMoves::~BestMoves()
{
	delete[] Moves;
}

  void BestMoves::Reset()
{
	storedMoves = 0;
}

/***************************************************************************
 * Operations
 **************************************************************************/

void BestMoves::Clear()
{
	if ( pthis == NULL ) BestMoves::Create();
	for ( int i = 0; i <= AISearchAgentMTDF::MaxSearchDepth; i++ ) 
		if ( pthis[ i ] != NULL )
			pthis[ i ]->Reset();
}

BestMoves* BestMoves::getInstance( int depth )
{
	if ( pthis == NULL ) BestMoves::Create();
	if ( pthis[ depth ] == NULL ) pthis[ depth ] = new BestMoves( depth );
	return pthis[ depth ];
}


void BestMoves::Create()
{
	pthis = ( BestMoves** )malloc( ( MaxSearchDepth + 1 ) * sizeof( BestMoves* ) );
	for ( int i = 0; i < ( MaxSearchDepth + 1 ); i++ )
		pthis[ i ] = NULL;
}

void BestMoves::Destroy()
{
	for ( int i = 0; i <= AISearchAgentMTDF::MaxSearchDepth; i++ ) 
		if ( pthis[ i ] != NULL ) delete pthis[ i ];
	free( pthis );
}


void BestMoves::addMove( int SourceField, int DestinationField, int Node, int Score, bool TranspositionHit )
{
	Moves[ depth-1 ].Node  = Node;
	Moves[ depth-1 ].Score = Score;
	
	Moves[ depth-1 ].SourceField      = SourceField;
	Moves[ depth-1 ].DestinationField = DestinationField;

	if ( TranspositionHit ) {
		storedMoves = 1;
		return;
	}

	if ( depth == 1 ) {
		storedMoves = 1;
	} else {
		memcpy( &Moves[ 0 ], &pthis[ depth-1 ]->Moves[ 0 ], ( depth-1 ) * sizeof( BestMoveEntry ) ); // optimisation: not copying cutoff moves

		storedMoves = pthis[ depth-1 ]->storedMoves + 1;
	}
}


// Get return a string of the full move sequence of AI
// stored moves in TranspositionTable are used to get the sequence
string BestMoves::Get( int depth, Board* theBoard, TranspositionTable* TransTable, TranspositionTable* TransTableBlack)
{
	Board myBoard;
	string rets = "";
	Move myMove;
	MoveListGenerator myMoveListGenerator;
	bool MaxNode = true;
	TranspositionTable* myTranspositionTable;
	
	myBoard.Clone( theBoard );

	for ( int i = 0; i < depth; i++ ) {
		if ( ( myBoard.CurrentPlayer == Player::SIDE_WHITE ) || ( TransTableBlack == NULL ) )
			myTranspositionTable = TransTable; else myTranspositionTable = TransTableBlack;
		if ( !myTranspositionTable->LookupBoard( &myBoard, &myMove ) ) break;
		myMoveListGenerator.ComputeLegalMoves( &myBoard );

		if ( !myBoard.ApplyMove( myMoveListGenerator.FindMoveForSquares( myMove.SourceSquare, myMove.DestinationSquare ) ) ) {
			rets += "*";
			break;
		};
		string s;
		Board::SquareToCoord( myMove.SourceSquare, myMove.DestinationSquare, &s );
		rets += s;
		if ( i != depth - 1 ) rets += " ";
	}

	return rets;
}




// PrintVerbose shows full move sequence of AI including the score
// stored moves in TranspositionTable are used to get the sequence
void BestMoves::PrintVerbose( int depth, Board* theBoard, TranspositionTable* TransTable, TranspositionTable* TransTableBlack )
{
	Board myBoard;
	Move myMove;
	MoveListGenerator myMoveListGenerator;
	bool MaxNode = true;
	
	TranspositionTable *myTranspositionTable;

	myBoard.Clone( theBoard );

	*Output::cout << endl << "Verbose print of AI-Moves:" << endl;
	if ( pthis[ depth ] != NULL ) {
		for ( int i = depth - 1; i >= depth - pthis[ depth ]->storedMoves; i-- ) {
			if ( ( myBoard.CurrentPlayer == Player::SIDE_WHITE ) || ( TransTableBlack == NULL ) )
				myTranspositionTable = TransTable; else myTranspositionTable = TransTableBlack;
			myTranspositionTable->LookupBoard( &myBoard, &myMove );
			myMoveListGenerator.ComputeLegalMoves( &myBoard );
			if ( !myBoard.ApplyMove( myMoveListGenerator.FindMoveForSquares( pthis[ depth ]->Moves[ i ].SourceField, pthis[ depth ]->Moves[ i ].DestinationField ) ) )  {
				cout << "Error in PrintVerbose !"<< endl;
				return;
			};
			string s;
			Board::SquareToCoord( pthis[ depth ]->Moves[ i ].SourceField, pthis[ depth ]->Moves[ i ].DestinationField, &s );
			*Output::cout << s << "  score: " << pthis[ depth ]->Moves[ i ].Score / 100.0 << "\t@ node: " << pthis[ depth ]->Moves[ i ].Node <<  endl;
		}
		if ( depth - pthis[ depth ]->storedMoves != 0 ) {
			string s;
			Board::SquareToCoord( myMove.SourceSquare, myMove.DestinationSquare, &s );
			*Output::cout << "^-: " << s << "  score: " << myMove.MoveEvaluation / 100.0 << "\tstored in TT @ node: " << myMove.MovingPiece << endl;
			for ( i = depth - 1 - pthis[ depth ]->storedMoves; i >= 0; i-- ) {
				if ( ( myBoard.CurrentPlayer == Player::SIDE_WHITE ) || ( TransTableBlack == NULL ) )
					myTranspositionTable = TransTable; else myTranspositionTable = TransTableBlack;
				if ( !myTranspositionTable->LookupBoard( &myBoard, &myMove ) ) break;
				myMoveListGenerator.ComputeLegalMoves( &myBoard );

				if ( !myBoard.ApplyMove( myMoveListGenerator.FindMoveForSquares( myMove.SourceSquare, myMove.DestinationSquare ) ) ) {
					*Output::cout << "TTable move not valid !" << endl;	
					break; 
				}
				Board::SquareToCoord( myMove.SourceSquare, myMove.DestinationSquare, &s );
				*Output::cout << "^-: " << s << "  score: " << myMove.MoveEvaluation / 100.0 << "\tstored in TT @ node: " << myMove.MovingPiece << endl;
			} 
		}
	}
}

// PrintVerboseWithNodes is similar to PrintVerbose except for showing the Node where the BestMove is found
// needs addMove calls in search algorithm
void BestMoves::PrintVerboseWithNodes( int depth, Board* theBoard, TranspositionTable* TransTable, TranspositionTable* TransTableBlack )
{
	Board myBoard;
	Move myMove;
	MoveListGenerator myMoveListGenerator;
	bool MaxNode = true;
	TranspositionTable* myTranspositionTable;

	myBoard.Clone( theBoard );

	*Output::cout << endl << "Verbose print of AI-Moves:" << endl;
	for ( int i = 0; i < depth; i++ ) {
		 
		// get the move out the TranspositionTable
		if ( ( myBoard.CurrentPlayer == Player::SIDE_WHITE ) || ( TransTableBlack == NULL ) )
			myTranspositionTable = TransTable; else myTranspositionTable = TransTableBlack;
		if ( !myTranspositionTable->LookupBoard( &myBoard, &myMove ) ) break;
		myMoveListGenerator.ComputeLegalMoves( &myBoard );
		
		// apply Move
		if ( !myBoard.ApplyMove( myMoveListGenerator.FindMoveForSquares( myMove.SourceSquare, myMove.DestinationSquare ) ) ) {
			*Output::cout << "TTable move not valid !" << endl;	
			break;
		};
		string s;
		// move -> stdout
		Board::SquareToCoord( myMove.SourceSquare, myMove.DestinationSquare, &s );
		*Output::cout << s << "  score: " << myMove.MoveEvaluation << endl;
	} 
	return;
}
