// BoardEvaluator.h: Schnittstelle fr die Klasse BoardEvaluator.
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_BOARDEVALUATOR_H__30BC286C_4365_485B_842A_26445A3D35C2__INCLUDED_ )
#define AFX_BOARDEVALUATOR_H__30BC286C_4365_485B_842A_26445A3D35C2__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

class BoardEvaluator  
{
/***************************************************************************
 * Static Constants
 **************************************************************************/
private:
	/**
	  * The "graininess" of the evaluation.  MTD( f ) works a lot faster if the
	  * evaluation is relatively coarse
	  */
	static const int GRAIN;

public:
	static const int EVALUATE_BADBISHOPS;
	static const int EVALUATE_BADKNIGHTS;
	static const int EVALUATE_COMPLETE;
	static const int EVALUATE_DEVELOPMENT;
	static const int EVALUATE_KINGTROPISM;
	static const int EVALUATE_MATERIAL;
	static const int EVALUATE_PAWNSTRUCTURE;
	static const int EVALUATE_ROOKBONUS;
	static const int EVALUATE_ARRAY_SIZE;
/***************************************************************************
 * Data Members
 **************************************************************************/
private:
	// Data counters to evaluate pawn structure
	int MaxPawnFileBins[ 8 ];
	int MaxPawnColorBins[ 2 ];
	int MaxTotalPawns;
	int PawnRams;
	int MaxMostAdvanced[ 8 ];
	int MaxPassedPawns[ 8 ];
	int MinPawnFileBins[ 8 ];
	int MinMostBackward[ 8 ];
  
/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
public:
	BoardEvaluator();
	virtual ~BoardEvaluator();
	
/***************************************************************************
 * Accessors
 **************************************************************************/
/***************************************************************************
 * Static Operations
 **************************************************************************/
public:
	static int GetGrain() { return GRAIN; };
/***************************************************************************
 * Operations
 **************************************************************************/
public:
	void PrintVerbose( Board *theBoard, int fromWhosePerspective, bool newBoard = false );
	/** 
	  * A simple, fast evaluation based exclusively on material.  Since material
	  * is overwhelmingly more important than anything else, we assume that if a
	  * position's material value is much lower ( or much higher ) than another,
	  * then there is no need to waste time on positional factors because they
	  * won't be enough to tip the scales the other way, so to speak.
	  */
	int EvaluateQuickie( Board *theBoard, int fromWhosePerspective, bool useGrain = true );
	/**
	  * A detailed evaluation function, taking into account several positional
	  * factors
	  */
	int EvaluateComplete( Board *theBoard, int fromWhosePerspective, bool useGrain = true );
	
private:
	/**
	  * All other things being equal, having your Knights, Queens and Rooks close
	  * to the opponent's king is a good thing.
	  * This method is a bit slow and dirty, but it gets the job done
	  */
	int EvalKingTropism( Board *theBoard, int fromWhosePerspective );
	/**
	 * Rooks are more effective on the seventh rank, on open files and behind
	 * passed pawns
	 */
	int EvalRookBonus( Board *theBoard, int fromWhosePerspective );
	/**
	 * Mostly useful in the opening, this term encourages the machine to move
	 * its bishops and knights into play, to control the center with its queen's
	 * and king's pawns, and to castle if the opponent has many major pieces on
	 * the board
	 */
	int EvalDevelopment( Board *theBoard, int fromWhosePerspective );
	/**
	 * If Max has too many pawns on squares of the color of his surviving bishops,
	 * the bishops may be limited in their movement
	 */
	int EvalBadBishops( Board *theBoard, int fromWhosePerspective );
	/**
	 * If one of Max's Knights stand at the 1st or 8th file or rank, it is
	 * limited in his movement
	*/
	int EvalBadKnights( Board *theBoard, int fromWhosePerspective );
	/**
	 * Given the pawn formations, penalize or bonify the position according to
	 * the features it contains.
	 */
	int EvalPawnStructure( int fromWhosePerspective );
	/**
	 * Look at pawn positions to be able to detect features such as doubled,
	 * isolated or passed pawns
	 */
	bool AnalyzePawnStructure( Board *theBoard, int fromWhosePerspective );

};

#endif // !defined( AFX_BOARDEVALUATOR_H__30BC286C_4365_485B_842A_26445A3D35C2__INCLUDED_ )
