// Game.cpp: Implementierung der Klasse Game.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "Game.h"

/***************************************************************************
 * Static Class members
 **************************************************************************/

const int Game::HUMAN = 0;
const int Game::AI = 1;
const int Game::HUMANWINBOARD = 2;
const int Game::AIWINBOARD = 3;
const string Game::Version = "V1.3";
const string Game::LastMoveFilename = "LastMove.bin";
Game* Game::pthis;


/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/

Game::Game( string openingBook )
{
	pthis = this;
	Openings = new OpeningBook( openingBook );

	GameBoard.StartingBoard();

	Players[ 0 ] = NULL;
	Players[ 1 ] = NULL;
	Exporter = NULL;

}


Game::Game( string openingBook, char *logfilename )
{
	pthis = this;
	// load an existing opening book
	Openings = new OpeningBook( openingBook );

	// reset board settings
	GameBoard.StartingBoard();

	Players[ 0 ] = NULL;
	Players[ 1 ] = NULL;
	Exporter = NULL;

	Exporter = new PGNExporter();
	if ( Exporter->Open( logfilename ) == false ) {
		cerr << "Can't create logfile '" << logfilename << "'! Logging canceled!";
		delete Exporter;
		Exporter = NULL;
	}
}



Game::~Game()
{
	if ( Exporter ) {
		string s;
		Exporter->Close();
		
		Exporter->GetFileName( &s );
		*Output::cout << "Game-Log written to " << s << "!" << endl;
	}

	delete Openings;
}


/***************************************************************************
 * Operations
 **************************************************************************/

string Game::GetPlayerName( const int which )
{
	return Players[ which ]->PlayerString;
}


bool Game::SetPlayer( int which, int type, char *name )
{
	if ( type == HUMAN )
		Players[ which ] = new PlayerHuman( which );
	else if ( type == AI )
		Players[ which ] = new PlayerAI( which, AISearchAgent::AISEARCH_NEGASCOUT, Openings );
	else if ( type == HUMANWINBOARD )
		Players[ which ] = new PlayerHumanWinBoard( which );
	else if ( type == AIWINBOARD )
		Players[ which ] = new PlayerAIWinBoard( which, AISearchAgent::AISEARCH_NEGASCOUT, Openings );
		
	Players[ which ]->PlayerString = name;
	
	return true;
}

bool Game::SetPlayerBlack( int type, char *name ) 
{
	if ( type != HUMAN && type != AI && type != HUMANWINBOARD && type != AIWINBOARD ) {
		cerr << "Unknown Player Type ( black ), use Game::HUMAN or Game::AI or Game::HUMANWINBOARD or Game::AIWINBOARD!" << endl;
		return false;
	}
	
	SetPlayer( Player::SIDE_BLACK, type, name );
	
	if ( Exporter )
		Exporter->SetBlackPlayer( name );

	return true;
}



bool Game::SetPlayerWhite( int type, char *name ) 
{
	if ( type != HUMAN && type != AI && type != HUMANWINBOARD && type != AIWINBOARD ) {
		cerr << "Unknown Player Type ( white ), use Game::HUMAN or Game::AIor Game::HUMANWINBOARD or Game::AIWINBOARD!" << endl;
		return false;
	}
		SetPlayer( Player::SIDE_WHITE, type, name );
	
	if ( Exporter ) 
		Exporter->SetWhitePlayer( name );

	return true;
}



bool Game::CheckProgramConsistency()
{
	if ( _heapchk() != _HEAPOK ) return false;
	if ( cin.sync() == EOF ) return false;
	return true;
}



bool Game::RunGame()
{
	Player *CurrentPlayer;
	Move Mov;
		
	// Check if both Player are initialized
/*	if ( Players[ 0 ] == NULL || Players[ 1 ] == NULL ) {
		cerr << "Players aren't initialized! Use SetPlayerWhite() or SetPlayerBlack()!" << endl;
		return false;
	}
*/

	ifstream is;
	is.open( "C:\\test.bin", ios::in | ios::binary );

	do {
		if ( CheckProgramConsistency() != true ) {
			*Output::cout << "\nTHE LAST MOVE CORRUPTED THE MEMORY ( CORRUPTED POINTER? )! Debug!\n\n";
			exit( -1 );
		}

		// Show the current game board
		GameBoard.Print();

		// Show current evaluation
		BoardEvaluator eval;
		
		*Output::cout << setiosflags( ios::showpos );
		*Output::cout << "SCORE from perspective white " << (eval.EvaluateComplete( &GameBoard, Player::SIDE_WHITE ) - eval.EvaluateComplete( &GameBoard, Player::SIDE_BLACK )) / 200.0 << endl;
		*Output::cout << resetiosflags( ios::showpos );


		// Ask the next player for a move
		CurrentPlayer = Players[ GameBoard.GetCurrentPlayer() ];
		
		CurrentPlayer->GetMove( &GameBoard, &Mov );

		*Output::cout << Players[ GameBoard.GetCurrentPlayer() ]->PlayerString << " selects move: " ;
		Mov.Print();

		// Change the state of the game accordingly
       if  ( ( Mov.MoveType != Move::MOVE_RESIGN ) &&
		     ( Mov.MoveType != Move::MOVE_STALEMATE ) &&
			 ( Mov.MoveType != Move::MOVE_CHECKMATE ) &&
		     ( GameBoard.GetCurrentPlayer() == Player::SIDE_WHITE ) )
			storeLastMove();

		GameBoard.ApplyMove( &Mov );

		// If enabled, save the move in a file
		if ( Exporter ) Exporter->AppendMove( &Mov );

		// Pause
		
		// Players[ !GameBoard.GetCurrentPlayer() ]->ShowStats();
		// GameBoard.ShowInfo();
		
		// cleaup
		GameBoard.ClearExtraKings(Player::SIDE_WHITE);
		GameBoard.ClearExtraKings(Player::SIDE_BLACK);
    } while( ( Mov.MoveType != Move::MOVE_RESIGN ) &&
		     ( Mov.MoveType != Move::MOVE_STALEMATE ) &&
			 ( Mov.MoveType != Move::MOVE_CHECKMATE ) );

    Exporter->Close();

	*Output::cout << "GAME OVER.  Thanks for playing!" << endl ;
	return true;
}


bool Game::SetupGameBoard( char *filename ){
	// Load Boardsituation
	
	if ( GameBoard.Load( filename, &GameBoard ) == false )
		return false;
		else
		*Output::cout << "Loaded Board from file \"" << filename << "\"." << endl;
	

	return true;
}

bool Game::restoreLastMove()
{
	if ( GameBoard.MoveNo == 1 ) {
		*Output::cout << "Nervous ? Tired ?? Stupid ???" << endl;
		return false;
	}
	
	return load( LastMoveFilename );
}

bool Game::storeLastMove()
{
	return save( LastMoveFilename );
}

bool Game::load( string Filename )
{
	if ( Filename == "" ) {
		*Output::cout << "Please enter a valid filename ( e.g.: 'load mygame' ) !" << endl;
		return false;
	}
	LastMove.clear();
	LastMove.open( Filename.c_str(), ios::in | ios::binary );
	if ( !LastMove ) {
		*Output::cout << "Couldn't open '" << Filename << "'! File doesn't exist ?" << endl;
		return false;
	}
	GameBoard.read( LastMove );
	Players[ 0 ]->read( LastMove );
	Players[ 1 ]->read( LastMove );
	*Output::cout << endl << "Restored board:" << endl;
	LastMove.close();
	GameBoard.Print();

	return true;
	
}

bool Game::save( string Filename )
{
	if ( Filename == "" ) {
		*Output::cout << "Please enter a valid filename ( e.g.: 'save mygame' ) !" << endl;
		return false;
	}
	LastMove.clear();
	LastMove.open( Filename.c_str(), ios::in | ios::binary );
	if ( LastMove.is_open() ) {
		LastMove.close();
		if ( Filename != LastMoveFilename ) {
			*Output::cout << "File: '" << Filename << "' already exists. Overwrite  ? ( yes / no ) ";
			string s;
			cin >> s;

			char ch; // skip '\n' of cin
			cin.get( ch );
			
			if ( s[ 0 ] != 'Y' && s[ 0 ] != 'y' ) return false;
		}
	}
	LastMove.clear();
	LastMove.open( Filename.c_str(), ios::out | ios::binary );
	GameBoard.write( LastMove );
	Players[ 0 ]->write( LastMove );
	Players[ 1 ]->write( LastMove );
	LastMove.close();

	return true;
}

Game* Game::GetInstance()
{
	return pthis;
}
