// MoveListGenerator.h: Schnittstelle fr die Klasse MoveListGenerator.
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_MOVELISTGENERATOR_H__7529369A_BCC2_4710_9C2A_768EF33B278E__INCLUDED_ )
#define AFX_MOVELISTGENERATOR_H__7529369A_BCC2_4710_9C2A_768EF33B278E__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000

#include "Move.h"


typedef vector<int> intvector;


class MoveListGenerator  
{

/***************************************************************************
 * Static Data Members
 **************************************************************************/
private:
	/**
	  * Look-up tables containing all possible moves from all
	  * possible squares, by piece type
      *
	  * Destination = KnightMoves[ Source ][ MoveNo ]
	  * Destination = RookMoves[ Source ][ Direction ][ MoveNo ]
	  * U might ask, why a rook has a direction? It speeds up the MoveGeneration!
	  * If there's a friendly piece, you do not need to look behind this piece.
	  */
	static vector<int> KnightMoves[ 64 ];
	static vector<int> KingMoves[ 64 ];
	static vector<intvector> BishopMoves[ 64 ];
	static vector<intvector> RookMoves[ 64 ];

/***************************************************************************
 * Data Members
 **************************************************************************/
public:
	/**
	  * The list of moves, implemented as a java collection class, namely the
	  * ArrayList ( dynamic array )
	  */ 
  
	vector<Move> Moves;
	vector<Move>::iterator MovesIt;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
public:
	MoveListGenerator();
	virtual ~MoveListGenerator();
/***************************************************************************
 * Accessors
 **************************************************************************/
public:
	vector<Move> *GetMoveList() { return &Moves; }
	int Size() const { return Moves.size(); }

/***************************************************************************
 * Static Operations
 **************************************************************************/
public:
	static void InitStatics();
	static void DeleteStatics();

/***************************************************************************
 * Operations
 **************************************************************************/
public:
	void Sort();
	/**
	  * Prepare an iterator for scanning through the list of moves
	  */
	void ResetIterator();
	/**
	  * Look for a specific move in the list; if it is there, return true
	  * This is used by the PlayerHuman object, to verify whether a move entered
	  * by the player is actually valid
	  */
	bool Find( Move *mov );
	/**
	  * look for a move from "source" to "dest" in the list
	  */
	Move *FindMoveForSquares( int source, int dest );
	/**
	  * Find the next move in the list, if any
	  */
	Move *Next();
	/**
	  * Look at the board received as a parameter, and build a list of legal
	  * moves which can be derived from it.  If there are no legal moves, or if
	  * one of the moves is a king capture ( which means that the opponent's
	  * previous move left the king in check, which is illegal ), return false.
	  */
	bool ComputeLegalMoves( Board *theBoard );
	/**
	  * Find only the moves which are relevant to quiescence search; i.e., captures
	  */
	bool ComputeQuiescenceMoves( Board *theBoard );
	/**
	  * Dump the move list to standard output, for debugging purposes
	  */
	bool isAttacking( Board* theBoard, int sourceField, int destField );
	void Print();

private:
	bool ComputeWhiteQueenMoves( Board *theBoard );
	bool ComputeWhiteKingMoves ( Board *theBoard );
	/**
	  * Receives an extra "pieceType" parameter, because the queen AND the rook
	  * need to use this function
	  */
	bool ComputeWhiteRookMoves  ( Board *theBoard, int pieceType );
	bool ComputeWhiteBishopMoves( Board *theBoard, int pieceType );
	bool ComputeWhiteKnightMoves( Board *theBoard );
	bool ComputeWhitePawnMoves  ( Board *theBoard );

	bool ComputeBlackQueenMoves ( Board *theBoard );
	bool ComputeBlackKingMoves  ( Board *theBoard );
	bool ComputeBlackRookMoves  ( Board *theBoard, int pieceType );
	bool ComputeBlackBishopMoves( Board *theBoard, int pieceType );
	bool ComputeBlackKnightMoves( Board *theBoard );
	bool ComputeBlackPawnMoves  ( Board *theBoard );

	
	
	// same for quiescence
	bool ComputeWhiteQuiescenceQueenMoves( Board *theBoard );
	bool ComputeWhiteQuiescenceKingMoves ( Board *theBoard );
	/**
	  * Receives an extra "pieceType" parameter, because the queen AND the rook
	  * need to use this function
	  */
	bool ComputeWhiteQuiescenceRookMoves  ( Board *theBoard, int pieceType );
	bool ComputeWhiteQuiescenceBishopMoves( Board *theBoard, int pieceType );
	bool ComputeWhiteQuiescenceKnightMoves( Board *theBoard );
	bool ComputeWhiteQuiescencePawnMoves  ( Board *theBoard );

	bool ComputeBlackQuiescenceQueenMoves ( Board *theBoard );
	bool ComputeBlackQuiescenceKingMoves  ( Board *theBoard );
	bool ComputeBlackQuiescenceRookMoves  ( Board *theBoard, int pieceType );
	bool ComputeBlackQuiescenceBishopMoves( Board *theBoard, int pieceType );
	bool ComputeBlackQuiescenceKnightMoves( Board *theBoard );
	bool ComputeBlackQuiescencePawnMoves  ( Board *theBoard );
};

#endif // !defined( AFX_MOVELISTGENERATOR_H__7529369A_BCC2_4710_9C2A_768EF33B278E__INCLUDED_ )
