// OpeningBook.cpp: Implementierung der Klasse OpeningBook.
//
//////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "OpeningBook.h"


/***************************************************************************
 * Static Class members
 **************************************************************************/
const int    OpeningBook::TABLE_SIZE    = 1024;
const __int8 OpeningBook::FILETAG_NONE  = 0;
const __int8 OpeningBook::FILETAG_EOF   = 1;
const __int8 OpeningBook::FILETAG_NEXT  = 2;
const __int8 OpeningBook::FILETAG_BLACK = 3;
bool OpeningBook::used_for_last_move;
int OpeningBook::variants;
int OpeningBook::variantno;


/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
OpeningBook::OpeningBook()
{
	WhiteTable = ( OpeningBookEntry** ) malloc( sizeof( OpeningBookEntry* ) * TABLE_SIZE );
	BlackTable = ( OpeningBookEntry** ) malloc( sizeof( OpeningBookEntry* ) * TABLE_SIZE );
	for ( int i = 0; i < TABLE_SIZE; i++ ) {
		WhiteTable[ i ] = NULL;
		BlackTable[ i ] = NULL;
	}
	maxcollision = 0;
	entries = 0;
}

OpeningBook::OpeningBook( string& fileName )
{
	WhiteTable = ( OpeningBookEntry** ) malloc( sizeof( OpeningBookEntry* ) * TABLE_SIZE );
	BlackTable = ( OpeningBookEntry** ) malloc( sizeof( OpeningBookEntry* ) * TABLE_SIZE );
	for ( int i = 0; i < TABLE_SIZE; i++ ) {
		WhiteTable[ i ] = NULL;
		BlackTable[ i ] = NULL;
	}
	maxcollision = 0;
	entries = 0;
	Load( fileName );
}


OpeningBook::~OpeningBook()
{
	OpeningBookEntry* entry;
	OpeningBookEntry* successor;
	for ( int i = 0; i < TABLE_SIZE; i++ ) {
		entry = WhiteTable[ i ];
		while ( entry != NULL ) {
			successor = entry->nextEntry; 
			delete entry;
			entry = successor;
		}

		entry = BlackTable[ i ];
		while ( entry != NULL ) {
			successor = entry->nextEntry; 
			delete entry;
			entry = successor;
		}
	}
	free( WhiteTable );
	free( BlackTable );
}



/***************************************************************************
 * Operations
 **************************************************************************/


/* Querying the table for a ready-made move to play.  Return Null if there
   is none  */
Move *OpeningBook::Query( Board *theBoard ) {
	OpeningBookEntry* EntryPointer;
	Move* entryMove;

	vector<OpeningBookEntry*> validEntries;

	int countgood = 0;
	int countneutral = 0;
	int counttotal= 0;
	int rnd;
	OpeningBook::used_for_last_move = false;
	OpeningBook::variantno  = -1;
	OpeningBook::variants   =  0;

	// Select table of current player
	if ( theBoard->CurrentPlayer == Player::SIDE_WHITE ) {
		EntryPointer       =  WhiteTable[ abs( theBoard->HashKey() % TABLE_SIZE ) ];
	} else {
		EntryPointer       =  BlackTable[ abs( theBoard->HashKey() % TABLE_SIZE ) ];
	}

	int theLock = theBoard->HashLock();

	while ( EntryPointer != NULL ) {
		entryMove = EntryPointer->openingMove;
		// If the hash lock doesn't match the one for our position, try next
		if ( EntryPointer->theLock == theLock ) {
		// If the hash lock match -> add move in vector EntryPointer
			countgood +=    EntryPointer->rating_good;
			countneutral += EntryPointer->rating_neutral;
			validEntries.push_back( EntryPointer );
		}
		EntryPointer = EntryPointer->nextEntry;
	}
	
    // if we haven't found anything useful, quit
	if ( validEntries.empty() ) return NULL;
	
	// algorithm to select a move out of several...
	counttotal = countgood * 2 + countneutral;

	while ( true ) {
		rnd = rand() % validEntries.size();
		double a = ( ( ( validEntries[ rnd ]->rating_good * 2 + validEntries[ rnd ]->rating_neutral ) / ( double ) counttotal ) );
		double b = ( ( rand() % 10000 ) / 10000.0 );
		if ( a > b ) {
			OpeningBook::used_for_last_move = true;
			OpeningBook::variantno = rnd + 1;
			OpeningBook::variants  = validEntries.size();
			return validEntries[ rnd ]->openingMove;
		}
	};


	return EntryPointer->openingMove;
}


/* adds a new PGN file to the opening book. Used for book creation. PGN must be in SAN format */
int OpeningBook::includePGN( string pgnFile ){
	// Open  pgnfile....
	ifstream sourceFile;
	sourceFile.open( pgnFile.c_str() );
	if ( !sourceFile ) {
		cerr << "Can't open source file ( " << pgnFile << " ) for creation...!" << endl;
		return 0;
	}

	int rating;
	Board openingBoard;
	MoveListGenerator openingMoves;
	
	//PGN Parser handles the selected PGN File
	PGNParser openingPGN( sourceFile );  
	Move* openingMove;
	
	bool error = false;
	// repeat for all games in the PGN file
	while ( openingPGN.nextMatch() ) {  
//		*Output::cout << "GameNo " << openingPGN.matchNo << ":";
//		if ( ( openingPGN.PlayerWhite != "" ) && ( openingPGN.PlayerBlack != "" ) )
//			*Output::cout << " " << openingPGN.PlayerWhite << " vs " << openingPGN.PlayerBlack;
//		*Output::cout << endl;
		openingBoard.StartingBoard();
		error = false;
		// a maximum search depth is selected to limit the size of the book
		// Moves are assigned until there are'nt any left ,the search depth is reached or an error occured
		while ( openingPGN.getNextMove( &openingMove, &rating ) && !error && ( openingPGN.moveNo <= max_creating_depth ) ) {
			// Hash Board....
			openingMoves.ComputeLegalMoves( &openingBoard );
			int theHash = openingBoard.HashKey();
			int theLock = openingBoard.HashLock();

			// create new OpeningBookEntry...
			AddMove( theHash, theLock, openingMove, rating, openingBoard.CurrentPlayer );
			
			// move the parsed move
			if ( !openingBoard.ApplyMove( openingMove ) ) {
				cerr << "Failure in OpeningBook::Create" << endl;
				error = true;
				exit( 0 );
			}
		}
	}

	sourceFile.close();
	*Output::cout << ">> Entries:       " << entries      << endl;
	*Output::cout << ">> Maxcollisions: " << maxcollision << endl;
	*Output::cout << ">> Size ( kb ):     " << ( ( entries * ( sizeof( Move ) + 1 ) )/1024 ) << endl << endl;

	return openingPGN.matchNo;
}


/* AddMove adds a new move to the opening book in memory */
void OpeningBook::AddMove( int theHash, int theLock, Move *newopeningMove, int rating, int currentPlayer ) {
	// poor moves are skipped
	if ( rating == Move::MOVE_RATING_POOR ) return;

	int countcollision = 0;

	OpeningBookEntry* EntryPointer;
	OpeningBookEntry** EntryPointerAdress;
	EntryPointerAdress = &EntryPointer;
	Move* entryMove;

	// there are two openingbooks - one for each side
	if ( currentPlayer == Player::SIDE_WHITE ) {
		EntryPointer       =  WhiteTable[ abs( theHash % TABLE_SIZE ) ];
		EntryPointerAdress = &WhiteTable[ abs( theHash % TABLE_SIZE ) ];
	} else {
		EntryPointer       =  BlackTable[ abs( theHash % TABLE_SIZE ) ];
		EntryPointerAdress = &BlackTable[ abs( theHash % TABLE_SIZE ) ];
	}

	// opening book uses a hastable which allows collisions
	// every OpeningBookEntry is connected with its follower

	// is there move already stored ?
	while ( EntryPointer != NULL ) {
		entryMove = EntryPointer->openingMove;
		if ( EntryPointer->theLock == theLock && entryMove->Equals( newopeningMove ) ) {
			// entry is already there... update statistics and return
			if ( rating == Move::MOVE_RATING_GOOD )    EntryPointer->rating_good    += 1;
			if ( rating == Move::MOVE_RATING_UNKNOWN ) EntryPointer->rating_neutral += 1;
			return;
		}
		countcollision++;
		if ( maxcollision < countcollision ) maxcollision = countcollision;
		EntryPointerAdress = &EntryPointer->nextEntry;
		EntryPointer = EntryPointer->nextEntry;
	}

	// no -> new Entry....
	entries++;

	EntryPointer = new OpeningBookEntry;
	entryMove = EntryPointer->openingMove;
	entryMove->Set( newopeningMove );
	
	// update the queue pointers ( new element at last position )
	*EntryPointerAdress = EntryPointer;
	EntryPointer->nextEntry = NULL;
	EntryPointer->theLock = theLock;

	if ( rating == Move::MOVE_RATING_GOOD )    EntryPointer->rating_good    += 1;
	if ( rating == Move::MOVE_RATING_UNKNOWN ) EntryPointer->rating_neutral += 1;
}

/* 'Load' opens a binary saved openingbook, including all MoveEntries */
bool OpeningBook::Load( string SourceFile ) {
	ifstream is;
	int table_size;
	__int8 tag;

	is.open( SourceFile.c_str(), ios::in | ios::binary );
	if ( !is ) {
		cerr << "Can't load OpeningBook!" << endl;
		return false;
	}

	// at first, you check wheather table_size has checked since OpeningBook creating
	// if true, book is not valid any more
	is.read( ( char* )&table_size, sizeof( table_size ) );

	if ( table_size != TABLE_SIZE ) {
		cerr << "TABLE_SIZE of file is not correct" << endl;
		return false;
	}
	
	OpeningBookEntry *entry;

	// read all entries
	is.read( ( char* )&tag, sizeof( tag ) );
	int i = 0;

	// FILETAG_BLACK: white book is finished
	// FILETAG_NEXT:  a white move is coming next
	// FILETAG_NONE:  the next hash entry
	// read white book until black book is found...
	while ( tag != FILETAG_BLACK ) {
		if ( tag == FILETAG_NEXT ) {
			i++;
		}
		if ( tag == FILETAG_NONE ) {
			entries++;
			entry = new OpeningBookEntry;
			entry->read( is );
			
			// pointers of the OpeningTable have to be updated
			OpeningBookEntry*  EntryPointer;
			OpeningBookEntry** EntryPointerAdress;
			
			EntryPointerAdress = &EntryPointer;

			EntryPointer       =  WhiteTable[ i ];
			EntryPointerAdress = &WhiteTable[ i ];

			while ( EntryPointer != NULL ) {
				EntryPointerAdress = &EntryPointer->nextEntry;
				EntryPointer       =  EntryPointer->nextEntry;
			}

			*EntryPointerAdress = entry;
		}
		is.read( ( char* )&tag, sizeof( tag ) );
	}

	i = 0;
	// read black book until end of file
	while ( tag != FILETAG_EOF ) {
		if ( tag == FILETAG_NEXT ) {
			i++;
		}
		if ( tag == FILETAG_NONE ) {
			entries++;
			entry = new OpeningBookEntry;
			entry->read( is );
			
			// pointers of the OpeningTable have to be updated
			OpeningBookEntry*  EntryPointer;
			OpeningBookEntry** EntryPointerAdress;
			
			EntryPointerAdress = &EntryPointer;

			EntryPointer       =  BlackTable[ i ];
			EntryPointerAdress = &BlackTable[ i ];

			while ( EntryPointer != NULL ) {
				EntryPointerAdress = &EntryPointer->nextEntry;
				EntryPointer       =  EntryPointer->nextEntry;
			}

			*EntryPointerAdress = entry;
		}

		is.read( ( char* )&tag, sizeof( tag ) );
	}
	is.close();

	*Output::cout << "OpeningsBook =       "  << setw( 7 ) << entries << " entries" << endl;
	return true;
}

/* Store saves the existing Opening book in memory to a file - all entries are stored binary */
bool OpeningBook::Store( string destFile ) {

	ofstream os;
	int table_size = TABLE_SIZE;

	os.open( destFile.c_str(), ios::out | ios::binary );
	if ( !os ) {
		cerr << "Can't open destfile for creating OpeningBook!" << endl;
		return false;
	}

	OpeningBookEntry *entry;
	
	// write TABLE_SIZE - for validiation at loading
	os.write( ( char* )&table_size, sizeof( table_size ) );

	int i;

	// every HashEntry of the OpeningTable is saved
	// WhiteBook:
	for ( i = 0; i < TABLE_SIZE; i++ ) {
		entry = WhiteTable[ i ];
		while ( entry != NULL ) {
			os.write( ( char* )&FILETAG_NONE, sizeof( FILETAG_NONE ) );
			entry->write( os );
			entry = entry->nextEntry; 
		}
		if ( ( i + 1 ) == TABLE_SIZE ) {
			os.write( ( char* )&FILETAG_BLACK, sizeof( FILETAG_BLACK ) );
		} else {
			os.write( ( char* )&FILETAG_NEXT, sizeof( FILETAG_NEXT ) );
		}
	}

	// BlackBook:
	for ( i = 0; i < TABLE_SIZE; i++ ) {
		entry = BlackTable[ i ];
		while ( entry != NULL ) {
			os.write( ( char* )&FILETAG_NONE, sizeof( FILETAG_NONE ) );
			entry->write( os );
			entry = entry->nextEntry; 
		}
		if ( ( i + 1 ) == TABLE_SIZE ) {
			os.write( ( char* )&FILETAG_EOF, sizeof( FILETAG_EOF ) );
		} else {
			os.write( ( char* )&FILETAG_NEXT, sizeof( FILETAG_NEXT ) );
		}
	}

	os.close();
	return true;
}

/* parameters of StartBookCreation are used to create new OpeningBook
   Have a look at the first line for syntax. */
bool OpeningBook::StartBookCreation( int argc, char* argv[] ) {
	// valid parameters ?
	int NumMatches = 0;

	if ( argc < 5 ) {
		cerr << "Not enough parameters for openingbookcreation !" << endl;
		cerr << "Parameter 1: \"createbook\"" << endl;
		cerr << "Parameter 2: <destination file>" << endl;
		cerr << "Parameter 3: <max_depth>" << endl;
		cerr << "Parameter 4..n: <source file>" << endl;

		return false;
	}

	// open new book in memory and include all PGN selected
	OpeningBook theBook;
	theBook.max_creating_depth = atoi( argv[ 3 ] );
	for ( int i = 4; i < argc; i++ ) {
		*Output::cout << "Open source file ( " << argv[ i ] << " ) for creation... " << endl;
		NumMatches += theBook.includePGN( argv[ i ] );
		*Output::cout << "Matches: " << NumMatches << endl;
	}


	// save new book to disk
	theBook.Store( argv[ 2 ] );

	return true;
}
