//
// Player.h - An abstract base class for all types of players
//
// Purpose: There are currently two types of players in JavaChess: the computer
// player and the human player.  At a later time, other types may be added,
// including the demo player ( which picks its moves from a file, in which a
// game has been recorded move by move ) and a network player ( an entity from
// which the game obtains moves via a socket connection ).
//
// History:
// 21.03.03 Created
//
// 
/////////////////////////////////////////////////////////////////////////////////

#if !defined( AFX_PLAYER_H__99D38F61_2089_4165_BC67_7343A4E91A8F__INCLUDED_ )
#define AFX_PLAYER_H__99D38F61_2089_4165_BC67_7343A4E91A8F__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


class Player  
{
 /***************************************************************************
  * Static Constants
  **************************************************************************/
public:
  
	/** black player */
	static const int SIDE_BLACK; 

	/** white player */
	static const int SIDE_WHITE;

	/** player names */
	string PlayerString;


/***************************************************************************
 * Data Members
 **************************************************************************/
private:	
	/** Data member: which side is this player representing? */
	int Side;

/***************************************************************************
 * Constructor / Destructor
 **************************************************************************/
public:
	Player();
	~Player();

/***************************************************************************
 * Accessors
 **************************************************************************/
public:
	/** accessor to Player.Side*/
	int GetSide() const { return Side; }
	/** accessor to Player.Side */
	bool SetSide( int s );
  
/***************************************************************************
 * Operations
 **************************************************************************/
public:
	/** Ask the player to provide a move, given the current board situation */
	virtual bool GetMove( Board* theBoard, Move* mov ) = 0;
	virtual istream& read ( istream& is ) const;
	virtual ostream& write( ostream& os ) const;
	virtual void ShowStats();
};



#endif // !defined( AFX_PLAYER_H__99D38F61_2089_4165_BC67_7343A4E91A8F__INCLUDED_ )
