// TranspositionEntry.h: Schnittstelle fr die Klasse TranspositionEntry.
//
//////////////////////////////////////////////////////////////////////

#if !defined( AFX_TRANSPOSITIONENTRY_H__763A68D9_640C_48FC_BB7C_071CA820BFB9__INCLUDED_ )
#define AFX_TRANSPOSITIONENTRY_H__763A68D9_640C_48FC_BB7C_071CA820BFB9__INCLUDED_

#if _MSC_VER > 1000
#pragma once
#endif // _MSC_VER > 1000


/***************************************************************************
 * A small internal class containing an AB value for a given position, and
 * a "hash lock" signature used to identify collisions between board positions
 * with the same basic hashing values.
 *
 * Note that there is no need to store the actual move leading to this value,
 * for two reasons: first, by the time we check on the transposition table, the
 * move has already been applied; second, our version of alphabeta only handles
 * moves themselves at the top level of the search, so this information would
 * be passed to non one!
 ***************************************************************************/
class TranspositionEntry  
{
public:
  static const int NULL_ENTRY;

  TranspositionEntry();
	virtual ~TranspositionEntry();

  // Data fields, beginning with the actual value of the board and whether this  
  // value represents an accurate evaluation or only a boundary
  int theEvalType;
  int theEval;

  // This value was obtained through a search to what depth?  0 means that
  // it was obtained during quiescence search ( which is always effectively
  // of infinite depth but only within the quiescence domain; full-width
  // search of depth 1 is still more valuable than whatever Qsearch result )
  int theDepth;

  // Board position signature, used to detect collisions
  int theLock;

  // What this entry stored so long ago that it may no longer be useful?
  // Without this, the table will slowly become clogged with old, deep search
  // results for positions with no chance of happening again, and new positions
  // ( specifically the 0-depth quiescence search positions ) will never be
  // stored!
  int timeStamp;
  int theSourceSquare;
  int theDestinationSquare;
  int theNode;

  istream& read( istream &is ) const;
  ostream& write( ostream &os ) const;

};

#endif // !defined( AFX_TRANSPOSITIONENTRY_H__763A68D9_640C_48FC_BB7C_071CA820BFB9__INCLUDED_ )
